//---------------------------------------------------------------------------
#include <vcl.h>
#pragma hdrstop
#include "CABArchiver.h"
#pragma package(smart_init)
#define FolderUp 900000
#define ProgramID 060
//---------------------------------------------------------------------------
// ValidCtrCheck is used to assure that the components created do not have
// any pure virtual functions.
//
//---------------------------------------------------------------------------
static inline void ValidCtrCheck(TCABArchiver *)
{
        new TCABArchiver(NULL);
}
namespace Cabarchiver
{
        void __fastcall PACKAGE Register()
        {
                 TComponentClass classes[1] = {__classid(TCABArchiver)};
                 RegisterComponents("MyControl1", classes, 0);
        }
}

//---------------------------------------------------------------------
//
//              The CABItem
//
//---------------------------------------------------------------------
__fastcall TCABItem::TCABItem()
{
     Extract = false;
}
void TCABItem::SetFileName(AnsiString file)
{
     FFileName = ExtractFileName(file);
     FileFull = file;
     FilePath = ExtractFileDir(file);
     Size = FileSize(file);
     UpFileName = MyUpper(FFileName);
     UpFilePath = MyUpper(FilePath);
     UpFileFull = MyUpper(FileFull);
     ReturnFileName = file;
     if (Extract){
         RecordFileName = RemoveSlash(FFileName);
     }
     else{
         if (ExtractFileDrive(ReturnFileName)>0){
             RecordFileName = RemoveSlash(Copy(ReturnFileName,4,-1));
         }
     }
     DoFileName = MyUpper(RecordFileName);
     IsSelected = false;
     if (ExtractFileDrive(file)>0)
          SearchFile = MyUpper(RemoveSlash(Copy(file,4,-1)));
     else SearchFile = MyUpper(RemoveSlash(file));
}
void TCABItem::SetCompression(TCompression comp)
{
     if (comp==None) FComp = MakeNoCompression();
     if (comp==MsZip) FComp = MakeMsZipCompression();
     if (comp==Lzx15) FComp = MakeLzxCompression(15);
     if (comp==Lzx16) FComp = MakeLzxCompression(16);
     if (comp==Lzx17) FComp = MakeLzxCompression(17);
     if (comp==Lzx18) FComp = MakeLzxCompression(18);
     if (comp==Lzx19) FComp = MakeLzxCompression(19);
     if (comp==Lzx20) FComp = MakeLzxCompression(20);
     if (comp==Lzx21) FComp = MakeLzxCompression(21);
}
//---------------------------------------------------------------------
//
//              The CABContents
//
//---------------------------------------------------------------------
__fastcall TCABContents::TCABContents(TCABArchiver* Owner)
        : TStringList()
{
      AOwner = Owner;
}
TCABItem* TCABContents::AddItem()
{
      TCABItem* New = new TCABItem();
      New->Extract = Extract;
      New->Comp = AOwner->Compression;
      AddObject(Count,(TObject*)New);
      return New;
}
TCABItem* TCABContents::GetItem(int index)
{
      return (TCABItem*)Objects[index];
}
double TCABContents::GetTotalSize()
{
      double Zero = 0;
      for (int i1=0;i1<Count;i1++){
           Zero = Zero + GetItem(i1)->Size;
      }
      return Zero;
}
int TCABContents::Search(AnsiString name)
{
      for (int i1=0;i1<Count;i1++){
          if (GetItem(i1)->SearchFile==MyUpper(name)) return i1;
      }
      return -1;
}
//---------------------------------------------------------------------
//
//              The Cab archiver...
//
//---------------------------------------------------------------------
__fastcall TCABArchiver::TCABArchiver(TComponent* Owner)
        : TComponent(Owner)
{
     Contents = new TCABContents(this);
     CabinetReader1 = new TCabinetReader(this);
     CabinetWriter1 = new TCabinetWriter(this);
     CabinetStream1 = new TStreamCabinetFileHandler(this);
     CabinetReader1->FileHandler = CabinetStream1;
     CabinetWriter1->FileHandler = CabinetStream1;
     CabinetReader1->OnCopyFile = CabinetReader1CopyFile;
     CabinetWriter1->FailOnIncompressible = false;
     CabinetWriter1->OnFileStatus = CabinetWriter1FileStatus;
     CabinetWriter1->OnFilePlacedEvent = CabinetWriter1FilePlacedEvent;
     CabinetWriter1->OnGetNextCabinet = CabinetWriter1GetNextCabinet;
     CabinetReader1->OnNextCabinet = CabinetReader1GetNextCabinet;
     Try = false;
}
__fastcall TCABArchiver::~TCABArchiver()
{
     try{
         delete Contents;
         delete CabinetReader1;
         delete CabinetWriter1;
         delete CabinetStream1;
     }
     catch(...){
     }

}
//---------------------------------------------------------------------------
void __fastcall TCABArchiver::SetExtract(bool extract)
{
     Contents->Extract = extract;
}
//---------------------------------------------------------------------------
bool __fastcall TCABArchiver::GetExtract()
{
     return Contents->Extract;
}
//---------------------------------------------------------------------------
void __fastcall TCABArchiver::CABMessage(int position, int total,AnsiString message,TStat mode,AnsiString Lapse)
{
   if (FOnCABMessage)  FOnCABMessage(0,position,total,message,mode,AnsiTime());
   Application->ProcessMessages();
}
bool __fastcall TCABArchiver::Compress()
{
    Start();
    Mode = cmCompress;
    CABMessage(0,0,"Preparation in run...",cmMessage);
    if (Contents->Count<=0) Raise(1);
    if (FileName=="") Raise(2);
    if (FMultiDisk) GetMultiDiskSize();
    CreateDirectory(ExtractFileDir(FileName));
    if (FMultiDisk) CabinetWriter1->Open(FileName,"Disk",Sesame,FolderUp,ProgramID);
    else CabinetWriter1->Open(FileName,"Disk",MaxInt,FolderUp,ProgramID);
    if (FMultiDisk) AddOptions<<afoExecuteOnExtract;
    else AddOptions>>afoExecuteOnExtract;
    Type = Contents->GetTotalSize();
    CABMessage(0,Type,MyFormat("%s in preparation",FileName),cmProgressFirst);
    Beta1 = 0;
    Beta2 = 0;
    CurrentFile = "";
//    AddOptions<<afoExecuteOnExtract;
    for (int i1=0;i1<Contents->Count;i1++){
          CurrentFile1 = 0;
          CurrentFile2 = Contents->GetItem(i1)->Size;
          CABMessage(CurrentFile1,CurrentFile2,Contents->GetItem(i1)->FileName,cmSecondFirst);
          CurrentFile = "Compress "+Contents->GetItem(i1)->FileName;
          CabinetWriter1->AddFile(Item[i1]->ReturnFileName,Item[i1]->RecordFileName,AddOptions,Item[i1]->FComp);
    }
    CABMessage(0,10,"",cmProgressFirst);
    CABMessage(10,0,"",cmProgress);
    CABMessage(10,0,"Flushing Buffers...",cmMessage);
    CabinetWriter1->FlushCabinet(true);
    CABMessage(10,0,"Closing Cabinet...",cmMessage);
    CabinetWriter1->Close();
    CABMessage(10,0,"Compression of "+FileName+" done!",cmMessage);
    End();
    return true;
}
void TCABArchiver::Raise(int Error)
{
    AnsiString Current1 = "TCABArchiver: ";
    switch(Error){
        case 1:throw Exception(Current1+"List is empty");
        case 2:throw Exception(Current1+"FileName is empty");
        case 3:throw Exception(Current1+"No extract directory select");
        case 4:throw Exception(Current1+"The Cab file doesn't exist");
        case 5:throw Exception(Current1+"The extraction is impossible!");
        case 6:throw Exception(Current1+"Impossible to create the ouput directory");
        case 7:throw Exception(Current1+"The spanning size is not correct!");
        case 8:throw Exception(Current1+"The spanning is not need!");
    };
}
bool TCABArchiver::ExtractAll()
{
    Start();
    CABMessage(0,0,"Preparation in run...",cmMessage);
    if (FileName=="") Raise(2);
    if (FExtractDir=="") Raise(3);
    if (!FileExist(FileName)) Raise(4);
    if (!DirectoryExist(FExtractDir)){
         CreateDirectory(FExtractDir);
         if (!DirectoryExist(FExtractDir)) Raise(6);
    }
    try{
        Total = GetContents();
        if (!FileDirExtraction){
           for (int i1=0;i1<Contents->Count;i1++){
                CreateDirectory(SlashSep(FExtractDir,Contents->GetItem(i1)->FilePath));
           }
        }
        Pos = 0;
        CABMessage(0,Total,"Preparing "+ExtractFileName(FileName),cmProgressFirst);
        Mode = cmExtractAll;
        CabinetReader1->ExtractFiles(FileName,FExtractDir,0);
        CABMessage(Total,Total,"Extract all files from "+ExtractFileName(FileName)+" done!",cmMessage);
    }
    catch(...){
        if (!Try) Raise(5);
    }
    End();
    return true;
}
void __fastcall TCABArchiver::CabinetReader1CopyFile(TObject *Sender,
       const AnsiString FileName, int UncompressedSize, short Date,
       short Time, short Attribs, TFileCopyAction &Action,
       int &DestFileHandle)
{
    if (Mode==cmExtractAll){
         CreateDirectory(ExtractFileDir(FileName));
         Pos = Pos + UncompressedSize;
         CABMessage(Pos,0,ExtractFileName(FileName),cmProgress);
         Action = fcaDefaultCopy;
         return;
    }
    if (Mode==cmGetContents){
         Sizer = Sizer + UncompressedSize;
         TCABItem* New = Contents->AddItem();
         New->FileName = FileName;
         Action = fcaSkip;
         return;
    }
    if (Mode==cmGetSizeOfList){
         if (Contents->Search(FileName)!=-1) Total = Total + UncompressedSize;
    }
    if (Mode==cmExtractFile){
         int Index = Contents->Search(FileName);
         if (Index==-1) return;
         if (!Contents->GetItem(Index)->IsSelected) return;
         CreateDirectory(ExtractFileDir(FileName));
         Pos = Pos + UncompressedSize;
         CABMessage(Pos,0,ExtractFileName(FileName),cmStatus);
         Action = fcaDefaultCopy;
         return;
    }
}
double TCABArchiver::GetContents()
{
    Sizer = 0;
    Contents->Clear();
    TMode Memoire = Mode;
    Mode = cmGetContents;
    CabinetReader1->ExtractFiles(FileName,FExtractDir,0);
    Mode = Memoire;
    return Sizer;
}
TCABItem* TCABArchiver::GetCabItem(int index)
{
    return Contents->GetItem(index);
}
bool TCABArchiver::ExtractFile()
{
    Start();
    CABMessage(0,0,"Preparation in run...",cmMessage);
    if (Contents->Count<=0) Raise(1);
    if (FileName=="") Raise(2);
    if (FExtractDir=="") Raise(3);
    Mode = cmExtractFile;
    if (!DirectoryExist(FExtractDir)){
         CreateDirectory(FExtractDir);
         if (!DirectoryExist(FExtractDir)) Raise(6);
    }
    try{
        if (!FileDirExtraction){
           for (int i1=0;i1<Contents->Count;i1++){
                if (Contents->GetItem(i1)->IsSelected)
                    CreateDirectory(SlashSep(FExtractDir,Contents->GetItem(i1)->FilePath));
           }
        }
        Pos = 0;
        Mode = cmGetSizeOfList;
        Total = 0;
        CabinetReader1->ExtractFiles(FileName,FExtractDir,0);
        CABMessage(0,Total,"Preparing "+ExtractFileName(FileName),cmMessage);
        CABMessage(0,Total,"",cmProgressFirst);
        Mode = cmExtractFile;
        CabinetReader1->ExtractFiles(FileName,FExtractDir,0);
        CABMessage(Total,Total,"Extract selected files from "+ExtractFileName(FileName)+ " done!",cmMessage);
   }
    catch(...){
        Raise(5);
    }
    End();
    return true;
}
void __fastcall TCABArchiver::CabinetWriter1FileStatus(TObject *Sender,int CompressedSize,int UncompressedSize,int &ResultCode)
{
    if (Mode==cmCompress){
       Beta1 = Beta1 + CompressedSize;
       Beta2 = Beta2 + UncompressedSize;
       CABMessage(Beta2,Type,CurrentFile,cmStatus);
       CurrentFile1 = CurrentFile1 + UncompressedSize;
       CABMessage(CurrentFile1,CurrentFile2,CurrentFile,cmSecond);
    }
    ResultCode = 1;
}
void __fastcall TCABArchiver::CabinetWriter1FilePlacedEvent(TObject* Sender,TCCAB &CabParameters,const AnsiString FileName,int FileLength,bool Continuation,bool &AbortProcessing)
{
    CurrentFile1 = 0;
    CABMessage(0,10,FileName,cmSecondFirst);
    CABMessage(10,0,FileName,cmSecond);
    AbortProcessing = false;
}
void TCABArchiver::Start()
{
    Begin = GetTickCount();
}
void TCABArchiver::End()
{
    Stop = GetTickCount()-Begin;
    CABMessage(0,10,"",cmProgressFirst,0);
    CABMessage(10,0,"",cmProgress,0);
    CABMessage(0,10,"",cmSecondFirst,0);
    CABMessage(10,0,"",cmSecond);
    CABMessage(10,0,"",cmTime,0);
}
AnsiString TCABArchiver::AnsiTime()
{
    int Lapse = GetTickCount()-Begin;
    int Reste = Lapse/1000;
    int Hour = Reste/3600;
    int Beta = Reste - Hour*3600;
    int Minute = Beta/60;
    int Seconde = Beta - Minute*60;
    return MyFormat("%.02dh%.02d.%.02d",Hour,Minute,Seconde);
}
void __fastcall TCABArchiver::CabinetWriter1GetNextCabinet(TObject* Sender,TCCAB &CabParameters,AnsiString &NewCabFileName,int PreviousCabEstimatedSize,bool &AbortProcessing)
{
    AnsiString Current1 = ExtractFileName(FileName);
    AnsiString Current2 = ExtractFileExt(FileName);
    AnsiString Current3 = Current1;
    if (Current2.Length()!=0) Current3 = Copy(Current1,1,Current1.Length()-Current2.Length());
    AnsiString Current4 = MyFormat("%s%d.CAB",Current3,CabParameters.iCab);
    AnsiString Current5 = ExtractFileDir(FileName);
    AnsiString Current6 = SlashSep(Current5,Current4);
    NewCabFileName = Current6;
    strcpy(CabParameters.szCab,ExtractFileName(NewCabFileName).c_str());
    strcpy(CabParameters.szCabPath,ExtractFilePath(NewCabFileName).c_str());
    CabParameters.iDisk = CabParameters.iCab;
}
void __fastcall TCABArchiver::GetMultiDiskSize()
{
    Sesame = 0;
    if (FTypeCut==InOctect) Sesame = FSize;
    if (FTypeCut==InKilobyte) Sesame = FSize*1024;
    if (FTypeCut==InMegabyte) Sesame = FSize*1024*1024;
    if (FSize<=0) Raise(7);
    if (FSize==MaxInt) Raise(8);
}
void __fastcall TCABArchiver::CabinetReader1GetNextCabinet(TObject* Sender,
    const AnsiString NextCabinetName,
    const AnsiString NextCabinetDisk,
    AnsiString &CabinetPath,
    TFDIERROR ErrorIndication, bool &Abort)
{
//    ShowMessage(NextCabinetName);
//    NextFileName = SlashSep(CabinetPath,NextCabinetName);
    Abort = false;
}
bool TCABArchiver::ExtractSpan()
{
    int Compteur = 0;
    Try = true;
    ExtractAll();
    bool Condition = true;
    AnsiString Memoire = FileName;
    while (Condition){
        AnsiString Current1 = ExtractFileName(Memoire);
        AnsiString Current2 = ExtractFileExt(Memoire);
        AnsiString Current3 = Current1;
        if (Current2.Length()!=0) Current3 = Copy(Current1,1,Current1.Length()-Current2.Length());
        Compteur++;
        AnsiString Current4 = MyFormat("%s%d.CAB",Current3,Compteur);
        AnsiString Current5 = ExtractFileDir(Memoire);
        AnsiString Current6 = SlashSep(Current5,Current4);
        if (!FileExist(Current6)) Condition = false;
        else{
            FileName = Current6;
            ExtractAll();
        }    
    }
    Try = false;
    return true;
}
